==> blcr-0.6.1.tar.gz: blcr-0.6.1/util/cr_restart/cr_restart.c <==
/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_restart.c,v 1.42 2007/08/01 22:05:55 phargrov Exp $
 */

#define _LARGEFILE64_SOURCE 1   /* For O_LARGEFILE */

#include <sys/time.h>
#include <sys/resource.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/fcntl.h>
#include <getopt.h>
#include <signal.h>
#include <sys/wait.h>
#include <sched.h>
#include <errno.h>
#include <pthread.h>

#ifndef O_LARGEFILE
  #define O_LARGEFILE 0
#endif

#include "libcr.h"
#include "blcr_config.h"

/* can't use argv[0] to get name, since libtool screws it up */
#define MY_NAME	"cr_restart"

static void usage(FILE *stream)
{
    fprintf(stream,
"Usage: " MY_NAME " checkpoint_file\n"
"Options:\n"
#if 0 /* XXX: These are not documented, due to questions about their usability/stability */
"     --stop          send SIGSTOP just before return to user space.\n"
"     --abort         send SIGABRT just before return to user space.\n"
"     --cont          send SIGCONT just before return to user space.\n"
"     --freeze        freeze restarted processes for a debugger.\n"
"                     (currently just an alias for --stop).\n"
#endif
"     --omit-maybe    use a heuristic to omit cr_restart from checkpoints (default)\n"
"     --omit-always   always omit cr_restart from checkpoints\n"
"     --omit-never    never omit cr_restart from checkpoints\n"
" -?, --help          print this help message.\n"
"     --version	      print version information.\n" 
    );
}

sig_atomic_t child = 0; /* who did we spawn? */
static void signal_child (int, siginfo_t *, void *);

static void signal_self(int sig)
{
    struct sigaction sa;

    /* restore default (in kernel) handler */
    sa.sa_handler = SIG_DFL;
    sa.sa_flags = SA_RESTART | SA_NOMASK;
    sigemptyset(&sa.sa_mask);
    (void)sigaction(sig, &sa, NULL);

    /* send to self */
    raise(sig);

    /* restore self as handler */
    sa.sa_sigaction = &signal_child;
    sa.sa_flags = SA_RESTART | SA_NOMASK | SA_SIGINFO;
    (void)sigaction(sig, &sa, NULL);
}

/* Best-effort to forward all signals to the restarted child
 * XXX: This is probably not complete in terms of special-cases
 */
static void signal_child (int sig, siginfo_t *siginfo, void *context)
{
    if ((siginfo->si_code > 0) &&	/* si_code > 0 indicates sent by kernel */
	(sig == SIGILL || sig == SIGFPE || sig == SIGBUS || sig == SIGSEGV )) {
	/* This signal is OUR error, so we don't forward */
	signal_self(sig);
    } else if (sig == SIGTSTP || sig == SIGTTIN || sig == SIGTTOU) {
	/* The catchable stop signals go to child AND self */
	(void)kill(child, sig);
	signal_self(sig);
    } else {
	/* default case */
	(void)kill(child, sig);
    }
}

static void print_version()
{
    printf(MY_NAME " version %s\n", CR_RELEASE_VERSION);
}

/* getopt enum */
enum {
    opt_version,
    opt_stop,
    opt_abort,
    opt_cont,
    opt_omit_maybe,
    opt_omit_always,
    opt_omit_never,
};

/* try to exit such that our exit code is same as our child */
static void mimic_exit(int status)
{
    if (WIFEXITED(status)) {
	/* easy to mimic normal return */
	exit(WEXITSTATUS(status));
    } else if (WIFSIGNALED(status)) {
	/* disable generation of a 'core' */
	struct rlimit r;
	r.rlim_cur = r.rlim_max = 0;
	(void)setrlimit(RLIMIT_CORE, &r);

	/* now raise the signal */
	signal_self(WTERMSIG(status));
    } else {
	fprintf(stderr, "Unexpected status from child\n");
	exit(-1);
    }
}

int child_main(void *arg) {
    int local_errno;
    int err;

    /* loop to create any required threads */
    if (arg) {
	struct cr_procs_tbl *proc = arg;
	int threads = proc->threads;
	int flags = proc->clone_flags;
	uintptr_t incr = sysconf(_SC_PAGESIZE);
	uintptr_t stack;
        int i;

	stack = (uintptr_t)mmap(0, threads * incr, PROT_READ | PROT_WRITE | PROT_EXEC,
				MAP_PRIVATE | MAP_ANONYMOUS, 0, 0);
	if ((void *)stack == MAP_FAILED) {
	    fprintf(stderr, "Failed to allocate stack space for child thread(s)\n");
	    exit(1);
	}
#if (CR_STACK_GROWTH < 0) /* Stack grows down */
	stack += incr;
#elif (CR_STACK_GROWTH > 0) /* Stack grows up */
	/* Nothing to do */
#else
	#error "Don't know which way the stack grows"
#endif

        for (i = 1; i < threads; ++i) {
	    err = clone(&child_main, (void *)stack, flags, NULL);
	    if (err < 0) {
	        perror("clone()");
	        exit(1);
	    }
	    stack += incr;
	}
    }
	    
    /* Now overlay ourselves with the new image */
    err = __cri_syscall(CR_OP_RSTRT_CHILD, CRI_SYSCALL_NOARG, &local_errno);

    /* Not reached unless we've encountered a fatal error.
     * ********** Things are hairy here **********
     * We may have an incomplete thread environment, so we can't use libc fully.
     */

    /* Don't even try to print an error message */

    /* Try to make the entire thread group exit, if supported */
    __cri_exit_group(local_errno, NULL);
    /* We might return here in the case of ENOSYS */

    /* Note that this is really _exit() */
    __cri_exit(local_errno, NULL);
    /* NOTREACHED */

    return 0;
}

static volatile enum {
	CR_RSTRT_OMIT_MAYBE = 0,
	CR_RSTRT_OMIT_ALWAYS,
	CR_RSTRT_OMIT_NEVER
} omit_self = CR_RSTRT_OMIT_MAYBE;

static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;

/* "my" pid (done this way due to LinuxThreads limitations */
static int mypid;

/* Callback to perform a critical section OR get us omitted from the checkpoint */
static int my_cb(void *arg) {
    int status = CR_CHECKPOINT_PERM_FAILURE;
    const struct cr_checkpoint_info *info = cr_get_checkpoint_info();

    switch (omit_self) {
	case CR_RSTRT_OMIT_ALWAYS:
	    status = CR_CHECKPOINT_OMIT;
	    break;

	case CR_RSTRT_OMIT_NEVER:
	    status = CR_CHECKPOINT_READY;
	    break;

	case CR_RSTRT_OMIT_MAYBE:
	    if ((info->target == mypid) || (getppid() == 1)) {
		/* I am "root" of the checkpoint or an orphan (child of init) */
		status = CR_CHECKPOINT_OMIT;
	    } else {
		status = CR_CHECKPOINT_READY;
	    }
	    break;
    }

    /* Use of a mutex here produces a critical section, if and only
     * if we are included in the checkpoint.
     */
    if (status == CR_CHECKPOINT_READY) pthread_mutex_lock(&lock);
    cr_checkpoint(status);
    if (status == CR_CHECKPOINT_READY) pthread_mutex_unlock(&lock);
    return 0;
}

int main(int argc, char **argv)
{
    struct cr_procs_tbl procs;
    struct cr_rstrt_args req;
    int context_fd;
    int err;
    char *context_filename;
    char error_message[255];
    struct sigaction sa;
    sigset_t sigmask, oldmask;
    int sig;
    int child_status;
    int signal = 0;
    cr_client_id_t client_id;
    cr_callback_id_t cb_id;

    char * shortflags = "h?";  /* 1 colon == requires argument */
    struct option longflags[] = {
	/* misc options: */
	{ "help",         no_argument,  0, '?' },
	{ "version",      no_argument,  0, opt_version },
	{ "stop",         no_argument,  0, opt_stop },
	{ "freeze",       no_argument,  0, opt_stop },
	{ "abort",        no_argument,  0, opt_abort },
	{ "cont",         no_argument,  0, opt_cont },
	{ "omit-maybe",   no_argument,  0, opt_omit_maybe },
	{ "omit-always",  no_argument,  0, opt_omit_always },
	{ "omit-never",   no_argument,  0, opt_omit_never },
	{ 0,	     0,		        0, 0  }
    };

    while (1) {
	int longindex = -1;
	int opt = getopt_long(argc, argv, shortflags, longflags, &longindex);

	if (opt == -1)
	    break;		    /* reached last option */

	switch (opt) {
	    case opt_stop:
		signal = SIGSTOP;
		break;
	    case opt_abort:
		signal = SIGABRT;
		break;
	    case opt_cont:
		signal = SIGCONT;
		break;
	    case opt_omit_maybe:
		omit_self = CR_RSTRT_OMIT_MAYBE;
		break;
	    case opt_omit_always:
		omit_self = CR_RSTRT_OMIT_ALWAYS;
		break;
	    case opt_omit_never:
		omit_self = CR_RSTRT_OMIT_NEVER;
		break;
	    case opt_version:
		print_version();
		exit(0);
	    case '?':
	    case 'h':
		usage(stdout);
		exit(0);
	    default:
		usage(stderr);
		exit(-1);
	}
    }

    if (argc - optind != 1) {
        usage(stderr);
        return -1;
    } else {
      context_filename = argv[optind];
    }

    /* Record our pid */
    mypid = getpid();

    /* ... connect to the kernel */
    client_id = cr_init();
    if (client_id < 0) {
        fprintf(stderr, "Failed cr_init(): %d\n", client_id);
        return client_id;
    }

    /* ... register callback to exclude us from the checkpoint */
    cb_id = cr_register_callback(&my_cb, NULL, CR_THREAD_CONTEXT);
    if (cb_id < 0) {
        fprintf(stderr, "Failed cr_regsiter_callback(): %d\n", cb_id);
        return cb_id;
    }

    /* ... open up the context file */
    context_fd = open(context_filename, O_RDONLY | O_LARGEFILE);
    if (context_fd < 0) {
        /* XXX:  Clean this up to be consistent */
        snprintf(error_message, sizeof(error_message), 
          "open(%s, O_RDONLY)", context_filename);
        perror(error_message);
        return context_fd;
    }

    /* ... initialize the request structure */
    req.cr_fd       = context_fd;
    req.cr_opts     = NULL;
    req.signal      = signal;

    /* START of critical section.
     * Since we don't (yet?) try to checkpoint the state associated with
     * a partially completed restart, we must exclude checkpoints between
     * CR_OP_RSTRT_REQ and CR_OP_RSTRT_REAP
     */
    pthread_mutex_lock(&lock);

    /* ... tell the kernel to go off and restart something */
    err = cri_syscall(CR_OP_RSTRT_REQ, (uintptr_t)&req);
    if (err < 0) {
        /* The restart request was unsuccessful.  */
        fprintf(stderr, "cri_syscall(CR_OP_RSTRT_REQ, &req): %s\n", cr_strerror(errno));
        return err;
    }

    /* ... fork child(ren) to contain the restarted task(s)
     * Note that CR_OP_RSTRT_PROCS might block if it cannot yet
     * be determined if more processes are needed.
     */ 
    while ((err = cri_syscall(CR_OP_RSTRT_PROCS, (uintptr_t)&procs)) > 0) {
        err = fork();
        if (err < 0) {
            /* The fork was unsuccessful.  */
            perror("fork()");
            return err; /* XXX: kill children!?! */
        } else if (err == 0) {
	    child_main(&procs);
	    /* NOTREACHED */
	}
    }
    if (err < 0) {
        fprintf(stderr, "cri_syscall(CR_OP_RSTRT_PROCS): %s\n", cr_strerror(errno));
        return err; /* XXX: kill children!?! */
    }

    /* ... wait forever for the restart to complete */
    do {
        err = cri_syscall(CR_OP_RSTRT_DONE, (uintptr_t)NULL);
    } while ((err < 0) && (errno == EINTR));
    if (err < 0) {
        /* There was a problem with the restart.  */
        fprintf(stderr, "cri_syscall(CR_OP_RSTRT_DONE, NULL): %s\n", cr_strerror(errno));
	return err;
    }

    /* ... prepare to forward signals */
    /* XXX: what should we use in place of _NSIG? */
    sigfillset(&sigmask);
    sigdelset(&sigmask, SIGSTOP);
    sigdelset(&sigmask, SIGKILL);
    sigdelset(&sigmask, SIGCHLD);
    sigprocmask(SIG_SETMASK, &sigmask, &oldmask);
    sa.sa_sigaction = &signal_child;
    sa.sa_flags = SA_RESTART | SA_NOMASK | SA_SIGINFO;
    sigemptyset(&sa.sa_mask);
    for (sig = 1; sig < _NSIG; ++sig) {
	if (sig == SIGCHLD) continue;
	if (sig == SIGKILL) continue;
	if (sig == SIGSTOP) continue;
	if (sig == CR_SIGNUM) continue;

	/* best-effort so we ignore failures */
	(void)sigaction(sig, &sa, NULL);
    }

    /* ... close the context file so a restarted cr_restart won't depend on it */
    /* XXX: How early can we do this safely?  No documented answer! */
    (void)close(context_fd);

    /* ... reap the checkpoint, collecting a child pid on success */
    err = cri_syscall(CR_OP_RSTRT_REAP, CRI_SYSCALL_NOARG);
    if (err < 0) {
        /* Reap returns the reson for the failure. */
	int save_errno = errno;
        fprintf(stderr, "Restart failed: %s\n", cr_strerror(errno));
	return save_errno;
    } else {
	child = (pid_t)err;
    }

    /* END of critical section */
    pthread_mutex_unlock(&lock);

    /* ... now we forward the signals */
    sigprocmask(SIG_SETMASK, &oldmask, NULL);

    /* XXX should have the option of a non-blocking case */
    do {
        /* wait for the child we've inherited */
        err = waitpid(child, &child_status, __WCLONE | __WALL);
    } while ((err < 0) && (errno == EINTR)); /* retry on non-fatal signals */
    /* XXX wait for additional children? */
    if (err < 0) {
        /* Unable to reap the child */
        perror("waitpid()");
	return err;
    }	

    mimic_exit(child_status);
    return -1; /* NOT REACHED */
}

==> blcr-0.6.2.tar.gz: blcr-0.6.2/util/cr_restart/cr_restart.c <==
/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_restart.c,v 1.42.20.2 2007/12/08 00:07:36 phargrov Exp $
 */

#define _LARGEFILE64_SOURCE 1   /* For O_LARGEFILE */

#include <sys/time.h>
#include <sys/resource.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/fcntl.h>
#include <getopt.h>
#include <signal.h>
#include <sys/wait.h>
#include <sched.h>
#include <errno.h>
#include <pthread.h>

#ifndef O_LARGEFILE
  #define O_LARGEFILE 0
#endif

#include "libcr.h"
#include "blcr_config.h"

/* can't use argv[0] to get name, since libtool screws it up */
#define MY_NAME	"cr_restart"

static void usage(FILE *stream)
{
    fprintf(stream,
"Usage: " MY_NAME " checkpoint_file\n"
"\n"
"General options:\n"
"  -?, --help          print this help message.\n"
"  -v, --version       print version information.\n" 
"\n"
"Options for signal sent to process(es) after restart:\n"
"      --run           no signal sent: continue execution (default).\n"
"  -S, --signal NUM    signal NUM sent to all processes/threads.\n"
"      --stop          SIGSTOP sent to all processes.\n"
"      --term          SIGTERM sent to all processes.\n"
"      --abort         SIGABRT sent to all processes.\n"
"      --kill          SIGKILL sent to all processes.\n"
"      --cont          SIGCONT sent to all processes.\n"
/* "      --freeze        freeze restarted processes for a debugger.\n" */
/* "                      (currently just an alias for --stop).\n" */
"\n"
"Options for checkpoints of restarted process(es):\n"
"      --omit-maybe    use a heuristic to omit cr_restart from checkpoints (default)\n"
"      --omit-always   always omit cr_restart from checkpoints\n"
"      --omit-never    never omit cr_restart from checkpoints\n"
    );
}

sig_atomic_t child = 0; /* who did we spawn? */
static void signal_child (int, siginfo_t *, void *);

static void signal_self(int sig)
{
    struct sigaction sa;

    /* restore default (in kernel) handler */
    sa.sa_handler = SIG_DFL;
    sa.sa_flags = SA_RESTART | SA_NOMASK;
    sigemptyset(&sa.sa_mask);
    (void)sigaction(sig, &sa, NULL);

    /* send to self */
    raise(sig);

    /* restore self as handler */
    sa.sa_sigaction = &signal_child;
    sa.sa_flags = SA_RESTART | SA_NOMASK | SA_SIGINFO;
    (void)sigaction(sig, &sa, NULL);
}

/* Best-effort to forward all signals to the restarted child
 * XXX: This is probably not complete in terms of special-cases
 */
static void signal_child (int sig, siginfo_t *siginfo, void *context)
{
    if ((siginfo->si_code > 0) &&	/* si_code > 0 indicates sent by kernel */
	(sig == SIGILL || sig == SIGFPE || sig == SIGBUS || sig == SIGSEGV )) {
	/* This signal is OUR error, so we don't forward */
	signal_self(sig);
    } else if (sig == SIGTSTP || sig == SIGTTIN || sig == SIGTTOU) {
	/* The catchable stop signals go to child AND self */
	(void)kill(child, sig);
	signal_self(sig);
    } else {
	/* default case */
	(void)kill(child, sig);
    }
}

static void die(int code, const char *format, ...)
		__attribute__ ((noreturn, format (printf, 2, 3)));
/*
 * die(code, format, args...)
 *
 * Sends an error message to stderr and exits program with indicated exit code.
 */
static void die(int code, const char *format, ...)
{
    va_list args;

    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);

    exit(code);
}

static void print_version()
{
    printf(MY_NAME " version %s\n", CR_RELEASE_VERSION);
}

static int
readint(const char *arg)
{
    int val;
    char *endptr;

    val = strtol(arg, &endptr, 10);
    if ((*arg == '\0') || (*endptr != '\0')) {
	usage(stderr);
	exit(-1);
    }

    return val;
}

/* getopt enum */
enum {
    opt_run = 1000,
    opt_stop,
    opt_term,
    opt_kill,
    opt_abort,
    opt_cont,
    opt_omit_maybe,
    opt_omit_always,
    opt_omit_never,
};

/* try to exit such that our exit code is same as our child */
static void mimic_exit(int status)
{
    if (WIFEXITED(status)) {
	/* easy to mimic normal return */
	exit(WEXITSTATUS(status));
    } else if (WIFSIGNALED(status)) {
	/* disable generation of a 'core' */
	struct rlimit r;
	r.rlim_cur = r.rlim_max = 0;
	(void)setrlimit(RLIMIT_CORE, &r);

	/* now raise the signal */
	signal_self(WTERMSIG(status));
    } else {
	fprintf(stderr, "Unexpected status from child\n");
	exit(-1);
    }
}

int child_main(void *arg) {
    int local_errno;
    int err;

    /* loop to create any required threads */
    if (arg) {
	struct cr_procs_tbl *proc = arg;
	int threads = proc->threads;
	int flags = proc->clone_flags;
	uintptr_t incr = sysconf(_SC_PAGESIZE);
	uintptr_t stack;
        int i;

	stack = (uintptr_t)mmap(0, threads * incr, PROT_READ | PROT_WRITE | PROT_EXEC,
				MAP_PRIVATE | MAP_ANONYMOUS, 0, 0);
	if ((void *)stack == MAP_FAILED) {
	    fprintf(stderr, "Failed to allocate stack space for child thread(s)\n");
	    exit(1);
	}
#if (CR_STACK_GROWTH < 0) /* Stack grows down */
	stack += incr;
#elif (CR_STACK_GROWTH > 0) /* Stack grows up */
	/* Nothing to do */
#else
	#error "Don't know which way the stack grows"
#endif

        for (i = 1; i < threads; ++i) {
	    err = clone(&child_main, (void *)stack, flags, NULL);
	    if (err < 0) {
	        perror("clone()");
	        exit(1);
	    }
	    stack += incr;
	}
    }
	    
    /* Now overlay ourselves with the new image */
    err = __cri_syscall(CR_OP_RSTRT_CHILD, CRI_SYSCALL_NOARG, &local_errno);

    /* Not reached unless we've encountered a fatal error.
     * ********** Things are hairy here **********
     * We may have an incomplete thread environment, so we can't use libc fully.
     */

    /* Don't even try to print an error message */

    /* Try to make the entire thread group exit, if supported */
    __cri_exit_group(local_errno, NULL);
    /* We might return here in the case of ENOSYS */

    /* Note that this is really _exit() */
    __cri_exit(local_errno, NULL);
    /* NOTREACHED */

    return 0;
}

static volatile enum {
	CR_RSTRT_OMIT_MAYBE = 0,
	CR_RSTRT_OMIT_ALWAYS,
	CR_RSTRT_OMIT_NEVER
} omit_self = CR_RSTRT_OMIT_MAYBE;

static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;

/* "my" pid (done this way due to LinuxThreads limitations */
static int mypid;

/* Callback to perform a critical section OR get us omitted from the checkpoint */
static int my_cb(void *arg) {
    int status = CR_CHECKPOINT_PERM_FAILURE;
    const struct cr_checkpoint_info *info = cr_get_checkpoint_info();

    switch (omit_self) {
	case CR_RSTRT_OMIT_ALWAYS:
	    status = CR_CHECKPOINT_OMIT;
	    break;

	case CR_RSTRT_OMIT_NEVER:
	    status = CR_CHECKPOINT_READY;
	    break;

	case CR_RSTRT_OMIT_MAYBE:
	    if ((info->target == mypid) || (getppid() == 1)) {
		/* I am "root" of the checkpoint or an orphan (child of init) */
		status = CR_CHECKPOINT_OMIT;
	    } else {
		status = CR_CHECKPOINT_READY;
	    }
	    break;
    }

    /* Use of a mutex here produces a critical section, if and only
     * if we are included in the checkpoint.
     */
    if (status == CR_CHECKPOINT_READY) pthread_mutex_lock(&lock);
    cr_checkpoint(status);
    if (status == CR_CHECKPOINT_READY) pthread_mutex_unlock(&lock);
    return 0;
}

int main(int argc, char **argv)
{
    struct cr_procs_tbl procs;
    struct cr_rstrt_args req;
    int context_fd;
    int err;
    char *context_filename;
    char error_message[255];
    struct sigaction sa;
    sigset_t sigmask, oldmask;
    int sig;
    int child_status;
    int signal = 0;
    cr_client_id_t client_id;
    cr_callback_id_t cb_id;

    char * shortflags = "vh?S:";  /* 1 colon == requires argument */
    struct option longflags[] = {
	/* misc options: */
	{ "help",         no_argument,  0, '?' },
	{ "version",      no_argument,  0, 'v' },
	/* signal options: */
	{ "signal",       required_argument, 0, 'S' },
	{ "run",          no_argument,  0, opt_run },
	{ "stop",         no_argument,  0, opt_stop },
	{ "freeze",       no_argument,  0, opt_stop },
	{ "term",         no_argument,  0, opt_term },
	{ "kill",         no_argument,  0, opt_kill },
	{ "abort",        no_argument,  0, opt_abort },
	{ "cont",         no_argument,  0, opt_cont },
	/* omit options: */
	{ "omit-maybe",   no_argument,  0, opt_omit_maybe },
	{ "omit-always",  no_argument,  0, opt_omit_always },
	{ "omit-never",   no_argument,  0, opt_omit_never },
	{ 0,	     0,		        0, 0  }
    };

    while (1) {
	int longindex = -1;
	int opt = getopt_long(argc, argv, shortflags, longflags, &longindex);

	if (opt == -1)
	    break;		    /* reached last option */

	switch (opt) {
	    /* Signal */
	    case 'S':
		signal = readint(optarg);
		if ((signal <= 0) || (signal >= _NSIG)) {
		    die(EINVAL,
			"Valid signal numbers are from 1 to %d, inclusive.\n",
			_NSIG - 1);
		}
		break;
	    case opt_run: /* --run */
		signal = 0;
		break;
	    case opt_stop: /* --stop */
		signal = SIGSTOP;
		break;
	    case opt_term: /* --term */
		signal = SIGTERM;
		break;
	    case opt_kill: /* --kill */
		signal = SIGKILL;
		break;
	    case opt_abort: /* --abort */
		signal = SIGABRT;
		break;
	    case opt_cont: /* --cont */
		signal = SIGCONT;
		break;
	    /* Omit */
	    case opt_omit_maybe:
		omit_self = CR_RSTRT_OMIT_MAYBE;
		break;
	    case opt_omit_always:
		omit_self = CR_RSTRT_OMIT_ALWAYS;
		break;
	    case opt_omit_never:
		omit_self = CR_RSTRT_OMIT_NEVER;
		break;
	    /* General */
	    case 'v':
		print_version();
		exit(0);
	    case '?':
	    case 'h':
		usage(stdout);
		exit(0);
	    default:
		usage(stderr);
		exit(-1);
	}
    }

    if (argc - optind != 1) {
        usage(stderr);
        return -1;
    } else {
      context_filename = argv[optind];
    }

    /* Record our pid */
    mypid = getpid();

    /* ... connect to the kernel */
    client_id = cr_init();
    if (client_id < 0) {
        fprintf(stderr, "Failed cr_init(): %d\n", client_id);
        return client_id;
    }

    /* ... register callback to exclude us from the checkpoint */
    cb_id = cr_register_callback(&my_cb, NULL, CR_THREAD_CONTEXT);
    if (cb_id < 0) {
        fprintf(stderr, "Failed cr_regsiter_callback(): %d\n", cb_id);
        return cb_id;
    }

    /* ... open up the context file */
    context_fd = open(context_filename, O_RDONLY | O_LARGEFILE);
    if (context_fd < 0) {
        /* XXX:  Clean this up to be consistent */
        snprintf(error_message, sizeof(error_message), 
          "open(%s, O_RDONLY)", context_filename);
        perror(error_message);
        return context_fd;
    }

    /* ... initialize the request structure */
    req.cr_fd       = context_fd;
    req.cr_opts     = NULL;
    req.signal      = signal;

    /* START of critical section.
     * Since we don't (yet?) try to checkpoint the state associated with
     * a partially completed restart, we must exclude checkpoints between
     * CR_OP_RSTRT_REQ and CR_OP_RSTRT_REAP
     */
    pthread_mutex_lock(&lock);

    /* ... tell the kernel to go off and restart something */
    err = cri_syscall(CR_OP_RSTRT_REQ, (uintptr_t)&req);
    if (err < 0) {
        /* The restart request was unsuccessful.  */
        fprintf(stderr, "cri_syscall(CR_OP_RSTRT_REQ, &req): %s\n", cr_strerror(errno));
        return err;
    }

    /* ... fork child(ren) to contain the restarted task(s)
     * Note that CR_OP_RSTRT_PROCS might block if it cannot yet
     * be determined if more processes are needed.
     */ 
    while ((err = cri_syscall(CR_OP_RSTRT_PROCS, (uintptr_t)&procs)) > 0) {
        err = fork();
        if (err < 0) {
            /* The fork was unsuccessful.  */
            perror("fork()");
            return err; /* XXX: kill children!?! */
        } else if (err == 0) {
	    child_main(&procs);
	    /* NOTREACHED */
	}
    }
    if (err < 0) {
        fprintf(stderr, "cri_syscall(CR_OP_RSTRT_PROCS): %s\n", cr_strerror(errno));
        return err; /* XXX: kill children!?! */
    }

    /* ... wait forever for the restart to complete */
    do {
        err = cri_syscall(CR_OP_RSTRT_DONE, (uintptr_t)NULL);
    } while ((err < 0) && (errno == EINTR));
    if (err < 0) {
        /* There was a problem with the restart.  */
        fprintf(stderr, "cri_syscall(CR_OP_RSTRT_DONE, NULL): %s\n", cr_strerror(errno));
	return err;
    }

    /* ... prepare to forward signals */
    /* XXX: what should we use in place of _NSIG? */
    sigfillset(&sigmask);
    sigdelset(&sigmask, SIGSTOP);
    sigdelset(&sigmask, SIGKILL);
    sigdelset(&sigmask, SIGCHLD);
    sigprocmask(SIG_SETMASK, &sigmask, &oldmask);
    sa.sa_sigaction = &signal_child;
    sa.sa_flags = SA_RESTART | SA_NOMASK | SA_SIGINFO;
    sigemptyset(&sa.sa_mask);
    for (sig = 1; sig < _NSIG; ++sig) {
	if (sig == SIGCHLD) continue;
	if (sig == SIGKILL) continue;
	if (sig == SIGSTOP) continue;
	if (sig == CR_SIGNUM) continue;

	/* best-effort so we ignore failures */
	(void)sigaction(sig, &sa, NULL);
    }

    /* ... close the context file so a restarted cr_restart won't depend on it */
    /* XXX: How early can we do this safely?  No documented answer! */
    (void)close(context_fd);

    /* ... reap the checkpoint, collecting a child pid on success */
    err = cri_syscall(CR_OP_RSTRT_REAP, CRI_SYSCALL_NOARG);
    if (err < 0) {
        /* Reap returns the reson for the failure. */
	int save_errno = errno;
        fprintf(stderr, "Restart failed: %s\n", cr_strerror(errno));
	return save_errno;
    } else {
	child = (pid_t)err;
    }

    /* END of critical section */
    pthread_mutex_unlock(&lock);

    /* ... now we forward the signals */
    sigprocmask(SIG_SETMASK, &oldmask, NULL);

    /* XXX should have the option of a non-blocking case */
    do {
        /* wait for the child we've inherited */
        err = waitpid(child, &child_status, __WCLONE | __WALL);
    } while ((err < 0) && (errno == EINTR)); /* retry on non-fatal signals */
    /* XXX wait for additional children? */
    if (err < 0) {
        /* Unable to reap the child */
        perror("waitpid()");
	return err;
    }	

    mimic_exit(child_status);
    return -1; /* NOT REACHED */
}
